import { h } from "@stencil/core";
import { ChartD3Service } from '../../services/chart.d3.service';
import { SeriesType } from '../../types/types';
import { ElementStrategyFactory, SingleLineSeries } from './chart.service';
import { ChartEvents } from './chart.events';
export class Chart {
    constructor() {
        this.inited = false;
        this.service = new SingleLineSeries();
    }
    componentWillLoad() {
        this.axisConfigs = {
            x: this.service.getDefaultXAxisConfig(),
            y: this.service.getDefaultYAxisConfig()
        };
    }
    async connectedCallback() {
        this.popover = this.element.querySelector('anj-overlay');
        this.popoverComponent = this.element.querySelector('[slot="popover"]');
        await this.initRender();
        this.inited = true;
    }
    disconnectedCallback() {
        this.popover = null;
        this.popoverComponent = null;
    }
    async onSeriesLoaded() {
        if (this.inited) {
            return this.initRender();
        }
    }
    onXAxisLoaded({ target, detail }) {
        const element = target;
        if (element.dimension === 'x') {
            this.xAxisElement = element;
            this.axisConfigs.x = Object.assign({}, this.axisConfigs.x, detail);
        }
        else {
            this.yAxisElement = element;
            this.axisConfigs.y = Object.assign({}, this.axisConfigs.y, detail);
        }
    }
    async handleResize() {
        this.clearChart();
        return this.debounceCreation();
    }
    // Add Component Events
    addLineEvents(svg, chartOverlay, x, data) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjHoverHandler(svg, data, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    addColumnEvents(svgBars, data) {
        svgBars
            .on('mousemove', (_d, i, bars) => this.eventsService.anjColumnHoverHandler(bars[i], data, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svgBars));
    }
    addStackedEvents(svg, bars) {
        bars
            .on('mousemove', (_d, i) => this.eventsService.anjStackedHoverHandler(svg, this.popoverData, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    // Creation & Rendering
    appendSVG(svg) {
        this.element.appendChild(svg.node());
    }
    async debounceCreation() {
        window.clearTimeout(this.creationTimer);
        return new Promise(resolve => {
            this.creationTimer = window.setTimeout(async () => {
                resolve(this.create());
            }, Chart.creationDebounceLength);
        });
    }
    async create() {
        const dims = this.service.getDims(this.element);
        // Establish the primary and secondary series
        this.primarySeries = (this.stackedColumns) ? SeriesType.Stacked : (this.lines && !this.columns) ? SeriesType.Line : SeriesType.Column;
        this.secondarySeries = ((this.stackedColumns || this.columns) && this.lines) ? SeriesType.Line : undefined;
        this.primaryComponents = (this.stackedColumns) ? this.stackedColumns : (this.lines && !this.columns) ? this.lines : this.columns;
        this.secondaryComponents = ((this.stackedColumns || this.columns) && this.lines) ? this.lines : undefined;
        // Establish the proper service strategy and manipulate the needed properties
        this.service = ElementStrategyFactory.get(this.primarySeries, this.primaryComponents);
        this.multiSeries = this.service.isMultiSeries();
        this.popover.placement = this.service.getPopoverPlacement();
        this.axisConfigs = this.service.xAxisTypeOverride(this.axisConfigs);
        // Configure Events Service for hover handling
        this.eventsService = new ChartEvents(this.primarySeries, this.anjHover, this.service, this.popover);
        // Fetch the various formats of data
        const { data, stackedData, minMax } = this.service.formatPrimaryData(this.axisConfigs);
        const flattenedData = this.service.flatten(data);
        // allows for primary and secondary data to be shown together
        this.popoverData = data;
        this.stackedData = stackedData;
        // Base for Chart
        const svg = ChartD3Service.createSVG(dims);
        // Attach Chart to the DOM
        this.appendSVG(svg);
        // Y-Axis
        const y = (this.yAxisElement)
            // Create defined x axis from anj-x-axis slot
            ? await this.hydrateYAxis(dims, flattenedData, svg, minMax)
            // Create hidden x axis for continuity
            : ChartD3Service.addYAxis(svg, dims, flattenedData, this.axisConfigs, minMax);
        // X-Axis
        const x = (this.xAxisElement)
            // Create defined x axis from anj-x-axis slot
            ? await this.hydrateXAxis(dims, flattenedData, svg, this.primarySeries, y)
            // Create hidden x axis for continuity
            : ChartD3Service.addXAxis(svg, dims, flattenedData, this.axisConfigs, undefined, this.primarySeries, y);
        // Components
        await this.hydratePrimaryComponents(svg, x, y, data, dims);
        await this.hydrateSecondaryComponents(svg, x, y, dims);
        return svg;
    }
    // Hydrate slotted components
    async hydratePrimaryComponents(svg, x, y, data, dims) {
        switch (this.primarySeries) {
            case SeriesType.Column:
                const x1 = ChartD3Service.getX1(x, data, this.multiSeries);
                return this.hydrateColumns(svg, data, x, x1, y, dims);
            case SeriesType.Stacked:
                return this.hydrateStackedColumns(svg, this.stackedData, x, y, dims);
            case SeriesType.Line:
                return this.hydrateLinesPrimary(svg, data, x, y, dims);
            default:
                return;
        }
    }
    async hydrateSecondaryComponents(svg, x, y, dims) {
        if (this.secondarySeries) {
            const lineData = this.secondaryComponents.map(l => l.data);
            const formattedLineData = this.service.coreFormatData(this.axisConfigs, lineData);
            return this.hydrateLinesSecondary(svg, formattedLineData, x, y, dims);
        }
        else {
            return Promise.resolve();
        }
    }
    async hydrateColumns(svg, data, x, x1, y, dims) {
        this.seriesMeta = [];
        const bars = await Promise.all(this.columns.map((c, i) => {
            this.seriesMeta[i] = { name: c.name, color: undefined };
            return c.init(svg, x, y, data[i], i, dims, x1);
        }));
        bars.forEach((b, i) => this.addColumnEvents(b, data[i]));
    }
    async hydrateLinesPrimary(svg, data, x, y, dims) {
        this.seriesMeta = [];
        if (this.lines.length === 1) {
            const renderArea = !data[0].some(point => point.y < 0);
            this.lines[0].area = renderArea;
        }
        const colors = await Promise.all(this.lines.map((l, i) => l.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries())));
        this.seriesMeta = colors.map((color, i) => {
            return { name: this.lines[i].name, color };
        });
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addLineEvents(svg, chartOverlay, x, data);
        // Animated L=>R reveal
        if (!this.service.isMultiSeries()) {
            ChartD3Service.addRevealAnimation(svg, dims, this.axisConfigs);
        }
    }
    async hydrateLinesSecondary(svg, data, x, y, dims) {
        const sIndex = this.primaryComponents.length;
        const colors = await Promise.all(this.lines.map((l, i) => l.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries(), sIndex)));
        colors.forEach((color, i) => {
            this.popoverData.push(data[i]);
            const popoverIndex = i + this.seriesMeta.length;
            this.seriesMeta[popoverIndex] = { name: this.lines[i].name, color };
        });
    }
    async hydrateStackedColumns(svg, data, x, y, dims) {
        this.seriesMeta = [];
        const rendered = await Promise.all(this.stackedColumns.map((c, i) => c.init(svg, x, y, data[i], i, dims)));
        // TODO: this should use the overlay
        this.seriesMeta = rendered.map(({ bars, color }, i) => {
            this.addStackedEvents(svg, bars);
            return { name: this.stackedColumns[i].name, color };
        });
    }
    hydrateXAxis(dims, data, svg, primarySeries, y) {
        return this.xAxisElement.init(svg, dims, data, this.axisConfigs, undefined, primarySeries, y);
    }
    hydrateYAxis(dims, data, svg, minMax) {
        return this.yAxisElement.init(svg, dims, data, this.axisConfigs, minMax, undefined);
    }
    // Load Slotted Components
    async initRender() {
        this.clearChart();
        const [columns, stackedColumns, lines] = await Promise.all([
            this.loadSeries('column'),
            this.loadSeries('stacked-column'),
            this.loadSeries('line')
        ]);
        this.columns = columns.length > 0 ? columns : undefined;
        this.stackedColumns = stackedColumns.length > 0 ? stackedColumns : undefined;
        this.lines = lines.length > 0 ? lines : undefined;
        return this.debounceCreation();
    }
    loadSeries(type) {
        return Promise.all(Array.from(this.element.querySelectorAll(`anj-series[type="${type}"]`))
            .map((col) => col.componentOnReady()));
    }
    clearChart() {
        const svg = this.element.querySelector('svg');
        if (svg) {
            svg.remove();
        }
    }
    render() {
        return [
            h("anj-overlay", { class: "anj-popover", placement: "top", offset: "0, 15", ref: r => this.popover = r },
                h("slot", { name: "popover" }),
                this.popoverComponent ? undefined : this.multiSeries && !this.columns ? h("anj-popover-content-multi", null) : h("anj-popover-content-single", null),
                (this.columns || !this.multiSeries) ? h("anj-arrow", null) : undefined)
        ];
    }
    static get is() { return "anj-chart"; }
    static get originalStyleUrls() { return {
        "$": ["chart.scss"]
    }; }
    static get styleUrls() { return {
        "$": ["chart.css"]
    }; }
    static get states() { return {
        "multiSeries": {},
        "popoverComponent": {}
    }; }
    static get events() { return [{
            "method": "anjHover",
            "name": "anjHover",
            "bubbles": true,
            "cancelable": true,
            "composed": true,
            "docs": {
                "tags": [],
                "text": "Emitted when the use hovers over a data point or series of them"
            },
            "complexType": {
                "original": "CoordinateData[]",
                "resolved": "CoordinateData[]",
                "references": {
                    "CoordinateData": {
                        "location": "import",
                        "path": "../../types/types"
                    }
                }
            }
        }]; }
    static get elementRef() { return "element"; }
    static get listeners() { return [{
            "name": "anjSeriesDidLoad",
            "method": "onSeriesLoaded",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "anjAxisDidLoad",
            "method": "onXAxisLoaded",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "resize",
            "method": "handleResize",
            "target": "window",
            "capture": false,
            "passive": true
        }]; }
}
Chart.creationDebounceLength = 100;
