import { ChartDataService } from '../../services/chart.data.service';
import { ChartD3Service } from '../../services/chart.d3.service';
import { SeriesStyle, SeriesType } from '../../types/types';
export class ElementStrategyFactory {
    static get(primaryType, primarySeries) {
        const data = primarySeries.map(c => c.data);
        switch (primaryType) {
            case SeriesType.Line:
                return (data.length > 1) ? new MultiLineSeries(data) : new SingleLineSeries(data);
            case SeriesType.Stacked:
                return new StackedSeries(data);
            case SeriesType.Column:
            default:
                return (data.length > 1) ? new MultiColumnSeries(data) : new SingleColumnSeries(data);
        }
    }
}
export class CoreFuncs {
    constructor(data) {
        this.data = data;
    }
    coreFormatData(axisConfigs, data) {
        const evalData = data || this.data;
        return (Array.isArray(evalData[0])) ? evalData.map(d => ChartDataService.formatData(axisConfigs, d)) : ChartDataService.formatData(axisConfigs, evalData);
    }
    formatPrimaryData(axisConfigs, data) {
        return { data: this.coreFormatData(axisConfigs, data) };
    }
    flatten(data) {
        return [].concat(...data);
    }
    getTargetingData(data) {
        return data[0];
    }
    xAxisTypeOverride(axisConfigs) {
        // If type is date and cannot coerce first x value to date, switch to string
        if (axisConfigs.x.labelType === 'date' && isNaN(Number(new Date(this.data[0][0].x)))) {
            axisConfigs.x.labelType = 'string';
        }
        return axisConfigs;
    }
    getDims(chartContainer) {
        const margin = { top: 20, right: 20, bottom: 20, left: 20 };
        const { height, width } = chartContainer.getBoundingClientRect();
        return {
            height: height - margin.top - margin.bottom,
            width: width - margin.left - margin.right,
            margin
        };
    }
    getDefaultXAxisConfig() {
        return {
            showAxis: false,
            labelType: 'date',
        };
    }
    getDefaultYAxisConfig() {
        return {
            showAxis: false,
            labelType: 'number',
        };
    }
}
export class ChartService {
}
export class SingleSeriesChartStrategy extends CoreFuncs {
    arrow() {
        return true;
    }
    getPopoverEl(svg, index) {
        return svg.select('.point-' + index).nodes()[0];
    }
    getPopoverPlacement() {
        return 'top';
    }
    getSeriesStyle() {
        return SeriesStyle.Single;
    }
    isMultiSeries() {
        return false;
    }
    getPopoverSeriesStyle() {
        return this.getSeriesStyle();
    }
}
export class MultiSeriesChartStrategy extends CoreFuncs {
    arrow() {
        return false;
    }
    getPopoverEl(svg, index) {
        return svg.select('.scrubber-' + index).nodes()[0];
    }
    getPopoverPlacement() {
        return 'right';
    }
    getSeriesStyle() {
        return SeriesStyle.Multi;
    }
    isMultiSeries() {
        return true;
    }
    getPopoverSeriesStyle() {
        return this.getSeriesStyle();
    }
}
export class StackedSeries extends MultiSeriesChartStrategy {
    formatPrimaryData(axisConfigs, data) {
        const baseFormat = this.coreFormatData(axisConfigs, data);
        const stackedData = ChartD3Service.formatDataAsStacked(baseFormat);
        const flattenedPointValues = stackedData.flat().flat();
        const minMax = { min: Math.min(...flattenedPointValues), max: Math.max(...flattenedPointValues) };
        return {
            data: baseFormat,
            stackedData,
            minMax
        };
    }
}
export class SingleLineSeries extends SingleSeriesChartStrategy {
    xAxisTypeOverride(axisConfigs) {
        axisConfigs.x.labelType = 'date';
        return axisConfigs;
    }
}
export class MultiLineSeries extends MultiSeriesChartStrategy {
    xAxisTypeOverride(axisConfigs) {
        axisConfigs.x.labelType = 'date';
        return axisConfigs;
    }
}
export class SingleColumnSeries extends SingleSeriesChartStrategy {
    getPopoverSeriesStyle() {
        return SeriesStyle.Single;
    }
}
export class MultiColumnSeries extends MultiSeriesChartStrategy {
    arrow() {
        return true;
    }
    getPopoverPlacement() {
        return 'top';
    }
    getPopoverSeriesStyle() {
        return SeriesStyle.Single;
    }
}
